"""glubdb.backends - implementation of actual db backends
"""

# TODO: MongoDB

# TODO: Other potential backends:
#       * The most popular NoSQL engines: Riak, Redis, Couch, Cassandra
#       * Memcache (and thus MemcacheDB)?
#       * Lucene/Solr
#       * Other Amazon or Google data storage? (SimpleDB, etc)
#       * should NOT recommend them, but what about Maria/MySQL and Postgres?
#       * File/pickle store?
#       * Python's dbm
